/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.storage.database;

import java.awt.*;
import java.sql.*;

import javax.swing.*;
import javax.swing.event.*;

import edu.stanford.smi.protege.util.*;

/**
 * TODO Class Comment
 * 
 * @author Ray Fergerson <fergerson@smi.stanford.edu>
 */
public class DatabaseWizardPage extends WizardPage {
    private JTextField driverField;
    private JTextField urlField;
    private JTextField tableField;
    private JTextField usernameField;
    private JTextField passwordField;
    private JTextArea errorArea;
    private DatabasePlugin plugin;

    public DatabaseWizardPage(Wizard wizard, DatabasePlugin plugin) {
        super("datbase", wizard);
        this.plugin = plugin;
        createComponents();
        layoutComponents();
        updateSetPageComplete();
    }

    private void createComponents() {
        driverField = createTextField();
        urlField = createTextField();
        tableField = createTextField();
        usernameField = createTextField();
        passwordField = createTextField();

        errorArea = ComponentFactory.createTextArea();
        errorArea.setEditable(false);
    }

    private JTextField createTextField() {
        JTextField field = ComponentFactory.createTextField();
        field.getDocument().addDocumentListener(new DocumentChangedListener() {
            public void stateChanged(ChangeEvent event) {
                updateSetPageComplete();
            }
        });
        return field;
    }

    private void updateSetPageComplete() {
        setErrorText(null);
        boolean isComplete = hasValidDriver()  && hasValidUrl() && hasValidTable();
        setPageComplete(isComplete);
    }

    private boolean hasValidDriver() {
        boolean isValid = false;
        String text = driverField.getText();
        if (text.length() == 0) {
            setErrorText("Driver class required");
        } else {
            Class clas = SystemUtilities.forName(text);
            isValid = clas != null;
            if (clas == null) {
                setErrorText("Driver class not found");
            }
        }
        return isValid;
    }

    private boolean hasValidUrl() {
        boolean isValid = false;
        String url = urlField.getText();
        String username = usernameField.getText();
        String password = passwordField.getText();
        if (url.length() == 0) {
            setErrorText("URL is required");
        } else if (username.length() == 0) {
            setErrorText("Username is required");
        } else {
	        try {
	            Connection c = DriverManager.getConnection(url, username, password);
	            c.close();
	            isValid = true;
	        } catch (SQLException e) {
	            setErrorText("Invalid URL, username, or password:\n" + e.getMessage());
	        }
        }
        return isValid;
    }

    private static int MAX_TABLE_NAME_LENGTH = 20;

    private boolean hasValidTable() {
        boolean isValid = false;
        String text = tableField.getText();
        if (text.length() == 0) {
            setErrorText("Table name is required.");
        } else if (text.length() > MAX_TABLE_NAME_LENGTH) {
            setErrorText("Table name is too long.");
        } else {
            isValid = true;
            for (int i = 0; i < text.length(); ++i) {
                char c = text.charAt(i);
                if (!Character.isJavaIdentifierPart(c)) {
                    isValid = false;
                    break;
                }
            }
            if (!isValid) {
                setErrorText("Invalid table name");
            }
        }
        return isValid;
    }

    private void setErrorText(String text) {
        errorArea.setText(text);
    }

    private void layoutComponents() {
        setLayout(new BorderLayout());
        Box panel = Box.createVerticalBox();
        addField(panel, driverField, "JDBC Driver Class Name");
        addField(panel, urlField, "JDBC Driver URL");
        addField(panel, tableField, "Table");
        addField(panel, usernameField, "Username");
        addField(panel, passwordField, "Password");
        
        errorArea.setPreferredSize(new Dimension(10, 50));
        errorArea.setBackground(getBackground());
        panel.add(Box.createVerticalStrut(20));
        panel.add(ComponentFactory.createScrollPane(errorArea));
        
        add(panel, BorderLayout.NORTH);
    }

    private void addField(Box panel, JComponent component, String text) {
        panel.add(new LabeledComponent(text, component));
    }

    public void onFinish() {
        plugin.setDriver(driverField.getText());
        plugin.setTable(tableField.getText());
        plugin.setUsername(usernameField.getText());
        plugin.setPassword(passwordField.getText());
        plugin.setURL(urlField.getText());
    }

}