/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.test;

import java.io.*;
import java.util.*;

import junit.framework.*;

import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.util.*;

/**
 * 
 * @author Ray Fergerson <fergerson@smi.stanford.edu>
 */
public abstract class AbstractTestCase extends TestCase {

    private static final File tempDirectory = createTempDirectory();

    private static File createTempDirectory() {
        File dir = null;
        try {
            File file = File.createTempFile("test", "tmp");
            File parent = file.getParentFile();
            file.delete();
            dir = new File(parent, "test" + File.separator);
            FileUtilities.deleteDirectory(dir);
            dir.mkdirs();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return dir;
    }

    protected static File getTempDirectory() {
        return tempDirectory;
    }

    protected static File getTempSubdirectory(String subDirectory) {
        File dir = new File(tempDirectory, subDirectory + File.separator);
        dir.mkdirs();
        return dir;
    }

    protected static void deleteTempSubdirectory(String subDirectory) {
        File dir = new File(tempDirectory, subDirectory);
        FileUtilities.deleteDirectory(dir);
    }

    protected static void checkErrors(Collection errors) {
        if (!errors.isEmpty()) {
            Iterator i = errors.iterator();
            while (i.hasNext()) {
                Object o = i.next();
                if (o instanceof Exception) {
                    Log.exception((Exception) o, APITestCase.class, "checkErrors");
                } else {
                    Log.warning(o.toString(), APITestCase.class, "checkErrors");
                }
            }
            fail();
        }
    }

    protected abstract Project getProject();

    protected KnowledgeBase getDomainKB() {
        return getProject().getKnowledgeBase();
    }

    protected static List makeList() {
        return new ArrayList();
    }

    public static List makeList(Object o) {
        assertNotNull(o);
        List list = makeList();
        list.add(o);
        return list;
    }

    public static List makeList(Object o1, Object o2) {
        assertNotNull(o2);
        List list = makeList(o1);
        list.add(o2);
        return list;
    }

    public static List makeList(Object o1, Object o2, Object o3) {
        assertNotNull(o3);
        List list = makeList(o1, o2);
        list.add(o3);
        return list;
    }

    public static List makeList(Object o1, Object o2, Object o3, Object o4) {
        assertNotNull(o4);
        List list = makeList(o1, o2, o3);
        list.add(o4);
        return list;
    }

    public static void assertEqualsList(Collection c1, Collection c2) {
        assertEquals(c1.size(), c2.size());
        Iterator i1 = c1.iterator();
        Iterator i2 = c2.iterator();
        while (i1.hasNext()) {
            Object o1 = i1.next();
            Object o2 = i2.next();
            assertEquals(o1, o2);
        }
    }

    public static void assertEqualsSet(Collection c1, Collection c2) {
        assertEquals(c1.size(), c2.size());
        Set s = new HashSet(c1);
        s.removeAll(c2);
        assertEquals(0, s.size());
    }

    protected Cls createCls() {
        return createCls(null);
    }

    protected Cls createCls(String name) {
        return createCls(name, getDomainKB().getRootCls());
    }

    protected Cls createCls(String name, Cls parent) {
        return getDomainKB().createCls(name, CollectionUtilities.createCollection(parent));
    }

    protected Cls createClsWithType(Cls type) {
        return getDomainKB().createCls(null, getDomainKB().getRootClses(), type);
    }

    protected Cls createSubClsWithType(Cls superclass, Cls type) {
        return getDomainKB().createCls(null, makeList(superclass), type);
    }

    protected Facet createFacet() {
        return getDomainKB().createFacet(null);
    }

    protected Frame createFrame() {
        return createCls();
    }

    protected void delete(Frame frame) {
        getDomainKB().deleteFrame(frame);
    }
    
    protected Instance createInstance(Cls cls) {
        return createInstance(null, cls);
    }

    protected Instance createInstance(String name, Cls cls) {
        Instance instance = getDomainKB().createInstance(name, cls);
        if (instance instanceof Cls) {
            ((Cls) instance).addDirectSuperclass(getDomainKB().getRootCls());
            // Log.trace("instance=" + instance, this, "createInstance", cls);
        }
        return instance;
    }

    protected Slot createSlot() {
        return getDomainKB().createSlot(null);
    }

    protected Slot createSlotOnCls(Cls cls) {
        Slot slot = createSlot();
        cls.addDirectTemplateSlot(slot);
        return slot;
    }

    protected Slot createMultiValuedSlot(ValueType type) {
        Slot slot = getDomainKB().createSlot(null);
        slot.setValueType(type);
        slot.setAllowsMultipleValues(true);
        return slot;
    }

    protected Slot createMultiValuedSlot(ValueType type, Cls cls) {
        Slot slot = createMultiValuedSlot(type);
        setCompleteValueType(cls, slot, type);
        return slot;
    }

    protected Slot createSingleValuedSlot(ValueType type) {
        Slot slot = getDomainKB().createSlot(null);
        slot.setValueType(type);
        slot.setAllowsMultipleValues(false);
        return slot;
    }

    protected Slot createSingleValuedSlot(ValueType type, Cls cls) {
        Slot slot = createSingleValuedSlot(type);
        setCompleteValueType(cls, slot, type);
        return slot;
    }

    protected Cls createSubCls(Cls parent) {
        return createCls(null, parent);
    }

    protected Slot createSubSlot(Slot parent) {
        return getDomainKB().createSlot(null, parent.getDirectType(), Collections.singleton(parent), true);
    }

    protected void deleteFrame(Frame frame) {
        getDomainKB().deleteFrame(frame);
    }

    protected Cls getCls(String name) {
        return getDomainKB().getCls(name);
    }

    protected Facet getFacet(String name) {
        return getDomainKB().getFacet(name);
    }

    public Frame getFrame(String name) {
        return getDomainKB().getFrame(name);
    }

    protected int getFrameCount() {
        return getDomainKB().getFrameCount();
    }

    protected Instance getInstance(String name) {
        return getDomainKB().getInstance(name);
    }

    protected void setCompleteValueType(Cls cls, Slot slot, ValueType type) {
        if (type == ValueType.INSTANCE) {
            slot.setAllowedClses(Collections.singleton(cls));
        } else if (type == ValueType.CLS) {
            slot.setAllowedParents(Collections.singleton(cls));
        } else {
            fail("bad type: " + type);
        }
    }

    public void assertEqualsArray(Object[] array1, Object[] array2) {
        assertEqualsList(Arrays.asList(array1), Arrays.asList(array2));
    }
}
